/*
 * Copyright (c) 2013,2014 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include <ivmob/ivmob.h>

#include "socketcan.h"
#include "ivmob-can.h"

static int16_t
decode_16(uint8_t *val)
{
	int16_t res;
	res = ((val[0] << 8) & 0xff00) |
	    (val[1] & 0x00ff);
	return res;
}

static uint16_t
decode_u16(uint8_t *val)
{
	uint16_t res;
	res = ((val[0] << 8) & 0xff00) |
	    (val[1] & 0x00ff);
	return res;
}

static uint32_t
decode_32(uint8_t *val)
{
	int32_t res;
	res = ((val[0] << 24) & 0xff000000) |
	    ((val[1] << 16) & 0x00ff0000) |
	    ((val[2] << 8) & 0x0000ff00) |
	    (val[3] & 0x000000ff);
	return res;
}

static uint32_t
decode_u32(uint8_t *val)
{
	uint32_t res;
	res = ((val[0] << 24) & 0xff000000) |
	    ((val[1] << 16) & 0x00ff0000) |
	    ((val[2] << 8) & 0x0000ff00) |
	    (val[3] & 0x000000ff);
	return res;
}

static void
decode_resp_speed(struct ivmob *ivmob, uint8_t *data)
{
	if (data[0] != 0 || data[1] != 0)
		warn("%s: send_speed problem %02x %02x\n",
		    __func__, data[0], data[1]);
}

static void
decode_processed_speed(struct ivmob *ivmob, uint8_t *data)
{
	ivmob->v = decode_16(data);
	ivmob->w = decode_16(data + 2);
	ivmob->cmd_origin = *(data + 4);
}

static void
decode_front_wheel_speed(struct ivmob *ivmob, uint8_t *data)
{
	ivmob->front_left_wheel_speed = decode_32(data);
	ivmob->front_right_wheel_speed = decode_32(data + 4);
}

static void
decode_rear_wheel_speed(struct ivmob *ivmob, uint8_t *data)
{
	ivmob->rear_left_wheel_speed = decode_32(data);
	ivmob->rear_right_wheel_speed = decode_32(data + 4);
}

static void
decode_front_current(struct ivmob *ivmob, uint8_t *data)
{
	ivmob->front_left_current = decode_32(data);
	ivmob->front_right_current = decode_32(data + 4);
}

static void
decode_rear_current(struct ivmob *ivmob, uint8_t *data)
{
	ivmob->rear_left_current = decode_32(data);
	ivmob->rear_right_current = decode_32(data + 4);
}

static void
decode_status(struct ivmob *ivmob, uint8_t *data)
{
	ivmob->temperature = *(data);
	ivmob->battery_level = *(data + 1);
	ivmob->mmi = *(data + 2);
	ivmob->safety_error_code = *(data + 3);
	ivmob->operational_mode = *(data + 4);
	ivmob->radio_buttons = *(data + 5);
	ivmob->radio_page = *(data + 6);
}

static void
decode_low_battery(struct ivmob *ivmob)
{
	printf("**** Low battery *****\n");
}

static void
decode_payload_off_rqst(struct ivmob *ivmob)
{
	printf("**** Payload off request *****\n");
	ivmobPayloadOffAck(ivmob);
}

/*
 * Decode a CAN message and update the ivmob structure
 */
static int
ivmobDecode(struct ivmob *ivmob, CAN_MSG *m)
{
	switch (m->id) {
	case IVMOB_RESP_SPEED_CMD:
		decode_resp_speed(ivmob, m->data);
		break;
	case IVMOB_RESP_ACTIVE_CMD:
		break;
	case IVMOB_RESP_INACTIVE_CMD:
		break;
	case IVMOB_RESP_LASER_SCANNER_CMD:
		break;
	case IVMOB_RESP_PROCESSED_SPEED:
		decode_processed_speed(ivmob, m->data);
		break;
	case IVMOB_RESP_FRONT_WHEEL_SPEED:
		decode_front_wheel_speed(ivmob, m->data);
		break;
	case IVMOB_RESP_REAR_WHEEL_SPEED:
		decode_rear_wheel_speed(ivmob, m->data);
		break;
	case IVMOB_RESP_FRONT_CURRENT:
		decode_front_current(ivmob, m->data);
		break;
	case IVMOB_RESP_REAR_CURRENT:
		decode_rear_current(ivmob, m->data);
		break;
	case IVMOB_RESP_STATUS:
		decode_status(ivmob, m->data);
		break;
	case IVMOB_RESP_SOFTWARE_VERSION:
		ivmob->version_major = m->data[0];
		ivmob->version_minor = m->data[1];
		ivmob->version_tiny = m->data[2];
		ivmob->version_build = m->data[3];
		break;
	case IVMOB_RESP_COMMISSIONING_DATE:
		ivmob->commissioning_year = m->data[2];
		ivmob->commissioning_month = m->data[1];
		ivmob->commissioning_day = m->data[0];
		break;
	case IVMOB_RESP_TOTAL_DISTANCE:
		ivmob->total_distance = decode_u32(m->data);
		break;
	case IVMOB_RESP_TOTAL_ON_TIME:
		ivmob->total_on_time = decode_u32(m->data);
		break;
	case IVMOB_RESP_WHEEL_SIZE:
		ivmob->wheel_diameter = decode_u16(m->data);
		break;
	case IVMOB_RESP_TORQUE_CONSTANT:
		ivmob->torque_constant = decode_u16(m->data);
		break;
	case IVMOB_RESP_GEAR_RATIO:
		ivmob->gear_ratio = *m->data;
		break;
	case IVMOB_RESP_WHEEL_TRACK_WIDTH:
		ivmob->wheel_track_width = decode_u16(m->data);
		break;
	case IVMOB_RESP_LASER_SCANNER:
		break;
	case IVMOB_RESP_PERIODIC_MSG:
		break;

	case IVMOB_PROCESSED_SPEED:
		decode_processed_speed(ivmob, m->data);
		break;
	case IVMOB_FRONT_WHEEL_SPEED:
		decode_front_wheel_speed(ivmob, m->data);
		break;
	case IVMOB_REAR_WHEEL_SPEED:
		decode_rear_wheel_speed(ivmob, m->data);
		break;
	case IVMOB_FRONT_CURRENT:
		decode_front_current(ivmob, m->data);
		break;
	case IVMOB_REAR_CURRENT:
		decode_rear_current(ivmob, m->data);
		break;
	case IVMOB_STATUS:
		decode_status(ivmob, m->data);
		break;

	case IVMOB_INVALID_COMMAND:
		break;

	case IVMOB_ALARM_SAFETY_MODE:
		ivmob->safety_error_code = m->data[0];
		break;
	case IVMOB_PAYLOAD_OFF_RQST:
		decode_payload_off_rqst(ivmob);
		break;
	case IVMOB_LOW_BATTERY:
		decode_low_battery(ivmob);
		break;

	default:
		return -1;
	}
	return 0;
}

/**
 * @brief Receive a CAN message and decode it
 */
int
ivmobReceiveAndDecode(struct ivmob *ivmob)
{
	struct CAN_MSG m;
	int rv;

	rv = socketcanReceiveMsgWait(ivmob->priv->socket, &m, 200);
	if (rv < 0)
		return rv;
	ivmob_lock(ivmob);
	if (ivmobDecode(ivmob, &m) != 0) {
		warnx("Can't decode frame: ");
		canMsgShow(stderr, &m);
		/* ignore error */
	}
	/* Signal waiting request */
	ivmob->priv->pending = m.id; /* CAN id of the response */
	pthread_cond_broadcast(&ivmob->priv->cond);
	ivmob_unlock(ivmob);
	return 0;
}

/*
 * Spawn a task reading data from the CAN controller and decoding messages
 * it also unlocks waiting requests functions
 */
void *
ivmobReadTask(void *arg)
{
	struct ivmob *ivmob = (struct ivmob *)arg;
	int rv;

	while (!ivmob->priv->done) {
		rv = ivmobReceiveAndDecode(ivmob);
		if (rv != 0 && errno == ETIMEDOUT)
			continue;
		if (rv != 0) {
			warn("ivmobReadTask: ivmobReceiveAndDecode");
			break;
		}
	}
	return NULL;
}
