/*
 * Copyright (c) 2013,2014 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <errno.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <ivmob/ivmob.h>

#include "socketcan.h"
#include "ivmob-can.h"

static void
encode_16(uint8_t *data, int16_t val)
{
	data[0] = (val & 0xff00) >> 8;
	data[1] = val & 0x00ff;
}

/* timespec addition - only works with positive values */
static void
timespec_add(struct timespec *a, const struct timespec *b)
{
	a->tv_sec += b->tv_sec;
	a->tv_nsec = a->tv_nsec + b->tv_nsec;
	if (a->tv_nsec > 999999999) {
		a->tv_sec += a->tv_nsec / 1000000000;
		a->tv_nsec = a->tv_nsec % 1000000000;
	}
}

/**
 * @brief Waits for the specified CAN ID message for timeo ms
 *
 * @param ivmob pointer to communication structure
 * @param rqst identifier of the response to wait for
 * @param timeo < 0 -> block
 * timeo = 0 -> no wait
 * timeo > 0 wait..
 *
 * @return 0 if message arrived; -1 otherwise;
 */
int
ivmobWait(struct ivmob *ivmob, int16_t rqst, int timeo)
{
	struct timespec d, ts;
	int rc;

	ivmob_lock(ivmob);
	if (timeo == 0) {
		if (ivmob->priv->pending == rqst) {
			ivmob->priv->pending = -1;
			ivmob_unlock(ivmob);
			return 0;
		}
	}
	if (timeo > 0) {
		d.tv_sec = timeo / 1000;
		d.tv_nsec = (timeo % 1000) * 1000000;
		clock_gettime(CLOCK_REALTIME, &ts);
		timespec_add(&ts, &d);
		rc = 0;
		/* wait for condition variable with timeout */
		while (ivmob->priv->pending != rqst && rc == 0)
			rc = pthread_cond_timedwait(&ivmob->priv->cond,
			    &ivmob->priv->mutex, &ts);
		errno = rc;
		ivmob->priv->pending = -1;
		ivmob_unlock(ivmob);
		return rc == 0 ? 0 : -1;
	}
	/* timeo < 0 -> blocking wait for condition variable */
	while (ivmob->priv->pending != rqst)
		pthread_cond_wait(&ivmob->priv->cond, &ivmob->priv->mutex);
	ivmob->priv->pending = -1;
	ivmob_unlock(ivmob);
	return 0;
}

/**
 * @brief Send a CAN message, and wait for the corresponding answer
 *
 * @param ivmob pointer to communication structure
 * @param rqst identifier of the response to wait for
 * @param timeo timeout see ivmobWait() for timeout values
 */
static int
ivmobSend(struct ivmob *ivmob, struct CAN_MSG *m, int16_t rqst, int timeo)
{
	int ret;

	ret = socketcanTransmitMsg(ivmob->priv->socket, m);
	if (timeo == 0 || ret < 0)
		return ret;
	return ivmobWait(ivmob, rqst, timeo);
}

/**
 * @brief Apply given speeds to the robot
 *
 * @param ivmob pointer to communication structure
 * @param v linear speed (mm/s)
 * @param w angular speed (mrad/s)
 * @param timeo timeout - see ivmobWait()
 * @return 0 (OK) or -1 (ERROR)
 */
int
ivmobCmdSpeed(struct ivmob *ivmob, int16_t v, int16_t w, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	encode_16(m.data, v);
	encode_16(m.data+2, w);
	m.len = 4;
	m.id = IVMOB_SPEED_CMD;
	return ivmobSend(ivmob, &m, IVMOB_RESP_SPEED_CMD, timeo);
}

/**
 * @brief Switch to active mode
 */
int
ivmobCmdActive(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_ACTIVE_CMD;
	return ivmobSend(ivmob,&m, IVMOB_RESP_ACTIVE_CMD, timeo);
}

/**
 * @brief Switch to inactive mode
 */
int
ivmobCmdInactive(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_INACTIVE_CMD;
	return ivmobSend(ivmob, &m, IVMOB_RESP_INACTIVE_CMD, timeo);
}

/**
 * @brief requests the last received speed commands
 */
int
ivmobRqstProcessedSpeed(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_PROCESSED_SPEED_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_PROCESSED_SPEED, timeo);
}

/**
 * @brief requests speed of front wheels
 */
int
ivmobRqstFrontWheelSpeed(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_FRONT_WHEEL_SPEED_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_FRONT_WHEEL_SPEED, timeo);
}

/**
 * @brief requests speed of rear wheels
 */
int
ivmobRqstRearWheelSpeed(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_REAR_WHEEL_SPEED_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_REAR_WHEEL_SPEED, timeo);
}

/**
 * @brief requests current drawn by front wheels
 */
int
ivmobRqstFrontCurrent(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_FRONT_CURRENT_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_FRONT_CURRENT, timeo);
}

/**
 * @brief requests current drawn by rear wheels
 */
int
ivmobRqstRearCurrent(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_REAR_CURRENT_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_REAR_CURRENT, timeo);
}

/**
 * @brief requests the robot status
 */
int
ivmobRqstStatus(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_STATUS_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_STATUS, timeo);
}

/**
 * @brief requests the robot firmware version
 */
int
ivmobRqstSoftwareVersion(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_SOFTWARE_VERSION_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_SOFTWARE_VERSION, timeo);
}

/**
 * @brief requests the robot first power on date
 */
int
ivmobRqstCommissioningDate(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_COMMISSIONING_DATE_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_COMMISSIONING_DATE, timeo);
}

/**
 * @brief requests the total distance run by the robot
 */
int
ivmobRqstTotalDistance(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_TOTAL_DISTANCE_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_TOTAL_DISTANCE, timeo);
}

/**
 * @brief requests the total power on time of the robot
 */
int
ivmobRqstTotalOnTime(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_TOTAL_ON_TIME_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_TOTAL_ON_TIME, timeo);
}

/**
 * @brief requests robot's wheels dimensions
 */
int
ivmobRqstWheelSize(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_WHEEL_SIZE_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_WHEEL_SIZE, timeo);
}

/**
 * @brief requests the robot's torque constant
 */
int
ivmobRqstTorqueConstant(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_TORQUE_CONSTANT_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_TORQUE_CONSTANT, timeo);
}

/**
 * @brief requests the robot's gear ratio
 */
int
ivmobRqstGearRatio(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_GEAR_RATIO_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_GEAR_RATIO, timeo);
}

/**
 * @brief requests the robot's track width
 */
int
ivmobRqstWheelTrackWidth(struct ivmob *ivmob, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_WHEEL_TRACK_WIDTH_RQST;
	return ivmobSend(ivmob, &m, IVMOB_RESP_WHEEL_TRACK_WIDTH, timeo);
}

/**
 * @brief requests periodic data messages from the robot
 *
 * @param ivmob pointer to communication structure
 * @param status_period
 * @param rear_current_period
 * @param front_current_period
 * @param rear_wheel_speed_period
 * @param front_wheel_speed_period
 * @param processed_speed_period
 * @param timeo timeout (ms)
 *
 * all periods are exprimed in 50ms increments.
 * 0 means no periodic transmission
 *
 * @return 0 (OK) or -1 (ERROR)
 */
int
ivmobRqstPeriodicMessages(struct ivmob *ivmob, uint8_t status_period,
    uint8_t rear_current_period, uint8_t front_current_period,
    uint8_t rear_wheel_speed_period, uint8_t front_wheel_speed_period,
    uint8_t processed_speed_period, int timeo)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 6;
	m.data[0] = status_period;
	m.data[1] = rear_current_period;
	m.data[2] = front_current_period;
	m.data[3] = rear_wheel_speed_period;
	m.data[4] = front_wheel_speed_period;
	m.data[5] = processed_speed_period;
	m.id = IVMOB_PERIODIC_MSG_CONFIG;
	return ivmobSend(ivmob, &m, IVMOB_RESP_PERIODIC_MSG, timeo);
}

/**
 * @brief Send the ack to a payload off message sent by the robot
 */
int
ivmobPayloadOffAck(struct ivmob *ivmob)
{
	CAN_MSG m;

	memset(&m, 0, sizeof(m));
	m.len = 0;
	m.id = IVMOB_PAYLOAD_OFF_CONFIRM;
	return socketcanTransmitMsg(ivmob->priv->socket, &m);
}
