/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <ivmob/ivmob.h>

#include "socketcan.h"
#include "ivmob-can.h"

/**
 * @brief Initialize communication with the 4MOB robot
 *
 * @param deviceName name of the socketcan interface connected
 *                        to the robot
 * @param readTask true - launch a thread reading and updating the ivmob
 *                  structure automatically
 *           false - the application will read the CAN bus,
 *                   decode the messages
 *                   and update the ivmob structure by itself
 *                   by calling ivmobReceiveAndDecode() periodically
 * @return pointer to an initialized ivmob structure or NULL
 */
struct ivmob *
ivmobInit(const char *deviceName, bool readTask)
{
	struct ivmob *ivmob;

	ivmob = (struct ivmob *)calloc(1, sizeof(struct ivmob));
	if (ivmob == NULL)
		return NULL;
	ivmob->priv = (struct ivmob_private *)
	    calloc(1, sizeof(struct ivmob_private));
	if (ivmob->priv == NULL) {
		free(ivmob);
		return NULL;
	}
	pthread_mutex_init(&ivmob->priv->mutex, NULL);
	pthread_cond_init(&ivmob->priv->cond, NULL);
	ivmob_lock(ivmob);
	ivmob->priv->socket = socketcanInit(deviceName);
	if (ivmob->priv->socket == -1) {
		ivmob_unlock(ivmob);
		free(ivmob->priv);
		free(ivmob);
		return NULL;
	}
	if (readTask) {
		if (pthread_create(&ivmob->priv->tid, NULL,
			ivmobReadTask, ivmob) != 0) {
			socketcanEnd(ivmob->priv->socket);
			ivmob_unlock(ivmob);
			free(ivmob->priv);
			free(ivmob);
			return NULL;
		}
	} else
		ivmob->priv->tid = -1;
	/* no request pending */
	ivmob->priv->pending = -1;
	ivmob_unlock(ivmob);
	return ivmob;
}

/**
 * @brief Stop communication with the robot and free data structures
 *
 * @param ivmob pointer to the communication structure to stop and free
 */
void
ivmobEnd(struct ivmob *ivmob)
{
	/* stop read task if needed */
	if (ivmob->priv->tid != -1) {
		ivmob->priv->done = 1;
		pthread_join(ivmob->priv->tid, NULL);
	}
	pthread_mutex_destroy(&ivmob->priv->mutex);
	pthread_cond_destroy(&ivmob->priv->cond);
	socketcanEnd(ivmob->priv->socket);
	free(ivmob->priv);
	free(ivmob);
}

/**
 * @brief Lock the communication structure
 */
int
ivmob_lock(struct ivmob *ivmob)
{
	return pthread_mutex_lock(&ivmob->priv->mutex);
}

/**
 * @brief Release the communication structure lock
 */
int
ivmob_unlock(struct ivmob *ivmob)
{
	return pthread_mutex_unlock(&ivmob->priv->mutex);
}
